#! /usr/bin/perl -w
use strict;


#=pod
my $user        = "zhanglabs";
my $datadir     = "/nfs/amino-home/zhanglabs/IonCom/output";
my $target      = "51";
my $model       = "protein.pdb";
my $bindir      = "/nfs/amino-home/zhanglabs/IonCom/bin/COACH/bin";
#=cut

=pod
$user        = "zhanglabs";
$datadir     = "/home/zhanglabs/COACH/output";
$target      = "CH000001";
$model       = "protein.pdb";
$bindir      = "/home/zhanglabs/COACH/bin";
=cut


my $inputdir  = "$datadir/$target";
if(!-s "$inputdir/$model")
{
    print "$model is missed\n";
    exit;
}

my $modelname = $model; $modelname=~s/\.pdb//;
my $outputdir = "$inputdir/$modelname";
my $tag       = "CL_$target\_$modelname";
my $CHoutdir = "$outputdir/coach";
`/bin/mkdir -p $CHoutdir`;

my $workdir  = "/tmp/$user/$tag";
`/bin/mkdir -p $workdir`;
`/bin/rm -rf $workdir/*`;
chdir "$workdir";
`cp $bindir/clustering .`;


my $cluster_cut=4.0;
my %queryCA = &load_Ca("$inputdir", $model);



my $cofdir = "$outputdir/cofactor";
my $cofinf = "selected_templates.lst";
my $all_templates = "templates_all.dat";
&cof_centers($cofdir, $cofinf, $all_templates);

my $tmsdir = "$outputdir/tmsite";
my $tmsinf = "templates.inf";
&tms_centers($tmsdir, $tmsinf, $all_templates);

my $sstdir = "$inputdir/ssite";
my $sstinf = "Bsites_fpt.clr";
&sst_centers($sstdir, $sstinf, $all_templates);

my $findir = "$outputdir/findsite";
my $fininf = "templates.inf";
&fin_centers($findir, $fininf, $all_templates);

my $condir = "$outputdir/concavity";
my $coninf = "Bsites_$modelname.dat";
&con_centers($condir, $coninf, $all_templates);

my @dis=();
my $n_selected = &dis_matrix($all_templates);

print "clustering...\n";
my @rst = `./clustering dis.dat $n_selected $cluster_cut`;
#print "@rst\n";
my @keys=keys %queryCA;
my $length=@keys;
$length = $length/3;

&vote(\@rst, $length, $all_templates, $n_selected);
print "$length\n";
&reRank();
`cp $all_templates $CHoutdir/`;
`cp Bsites* $CHoutdir/`;


#~~~~~~~~~~~~~~~~~~~~~~~ End procedure  ~~~~~~~~~~~~~~~~~~~~~~~~~#
print "Ending...\n";
my $time=`date`;
printf "ending time: $time";
`sync`;
`sync`;
sleep(1);


`rm -fr $workdir`;
print "Done!\n";


sub vote
{
    my ($rst_ref, $Lch, $inffile, $tot)=@_;

    my @rst=@$rst_ref;
    if(!-s "$inffile")
    {
	print "$inffile is missed\n";
	exit;
    }

    my @inf=`cat $inffile`;

    open(BOUT1, ">Bsites.dat");
    open(BOUT2, ">Bsites_prob.dat");
    open(BOUT3, ">Bsites.clr");

    my $n_cluster=0;
    foreach my $r (@rst)
    {
	chomp($r);
	if($r =~ /^cluster\s+(\d+):(.+)/)
	{
	    $n_cluster++;
	    my $c=$1;
	    my $line=$2;
	    my @a=split(/\s+/, $line);
	    my %q0=("COF"=>1.1, "TMS"=>0.57, "SST"=>0.5, "FIN"=>0.5, "CON"=>0.5);
	    my %q=%q0;
	    
	    #make consensus prediction in the cluster @a
	    my %prob=();
	    for(my $k=1; $k<=$Lch; $k++)
	    {
		$prob{$k}=0.0;
	    }
	    
	    my $score=0;
	    my $q0=-1;
	    my $pcount=0;
	    print BOUT3 "cluster $n_cluster\n";
	    foreach my $i(@a)
	    {
		$pcount++;		
		my $template = $inf[$i];
		chomp($template);
		my $PBS="";
		my $alg="";
		my $quality="";
		if($template =~ /(\S+).+\t(\S+)\s+:(\S+)/)
		{
		    $alg     = $1;
		    $quality = $2;
		    $PBS     = $3;
		}
		my @pred = split(/,/, $PBS);
		print BOUT3 "$template\n";
		
		if($q{$alg}<$quality)
		{
		    $q{$alg}=$quality;
		}

		for(my $k=0; $k<@pred; $k++)
		{
		    $prob{$pred[$k]} += 1;  #change here for weighted average
		}
	    }

	    for(my $k=1; $k<=$Lch; $k++)
	    {
		$prob{$k} /= $pcount if $pcount>0;
		printf BOUT2 "%.2f ", $prob{$k};
	    }
	    print BOUT2 "\n";
	    
	    
	    my @pres=();	    
	    my @keyr = sort{$prob{$b} <=> $prob{$a}} keys %prob;
	    my $count=0;
    
	    my $vcut=0.25;
	    foreach my $k(@keyr)
	    {   	
		if($prob{$k}>$vcut)
		{		
		    push(@pres, $k);
		} 
	    }
	    
	    while(@pres==0)
	    {	
		$vcut -=0.05;
		foreach my $k(@keyr)
		{   	
		    if($prob{$k}>$vcut)
		    {		
			push(@pres, $k);
		    } 
		}
	    }

	    my $qua=1;
	    my @method=keys %q0;
	    foreach my $m(@method)
	    {
		$qua = $qua * ($q{$m}/$q0{$m});
	    }	    
	    my $size = $pcount/$tot;
	    #my $cscore = $size * $qua;
	    my $cscore = 2/(1+exp(-$size*$qua)  ) -1;

	    
	    @keyr= sort {$a <=> $b} @pres;
	    my $ns=@a;
	    printf BOUT1 "%.2f\t%.2f\t%.2f\t:", $cscore, $size, $qua;	    
	    print BOUT1 join(",", @keyr) . "\n";
	}
    }
    close(BOUT1);
    close(BOUT2);
    close(BOUT3);    
}


sub reRank
{
    #re-rank Bsites.dat. Bsites,clr, Bsites_prob.dat based on c-score

    my %cscore=();
    my %pred=();
    my $i=0;
    
    open(IN, "Bsites.dat");
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /(\S+)\t(\S+)/)
	{
	    $cscore{$i}= $1;	    
	    $pred{$i}=$line;
	    $i++
	}
    }
    close(IN);

    $i=0;
    my %prob=();
    open(IN, "Bsites_prob.dat");
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /\d+/)
	{	    
	    $prob{$i}=$line;
	    $i++;
	}
    }
    close(IN);

    
    my %clr=();
    $i=-1;
    open(IN, "Bsites.clr");    
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /cluster/)
	{
	    $i++;
	}
	else
	{	 
	    push(@{$clr{$i}}, $line);
	}		
    }
    close(IN);

    my @keys= sort {$cscore{$b} <=> $cscore{$a}} keys %cscore;
    
    open(OUT, ">Bsites.dat");
    foreach my $k(@keys)
    {
	printf OUT "$pred{$k}\n";
    }
    close(OUT); 



    `cp $inputdir/$model query_tmp.pdb`;
    `cp $cofdir/lig_*.pdb .`;
    `cp $tmsdir/tms_*.pdb .`;
    `cp $findir/lig_*.pdb .`;

    my @pdbbuff=`cat query_tmp.pdb`;
    open(PDBT, ">query_tmp.pdb");
    foreach my $r(@pdbbuff)
    {
	chomp($r);
	print PDBT "$r\n";
    }
    print PDBT "TER\n";
    close(PDBT);


    $i=0;
    open(OUT1, ">Bsites.inf");
    open(OUT, ">Bsites.clr");
    foreach my $k(@keys)
    {
	my $conRes="";
	my $line=$pred{$k};
	if($line=~/.+:(\S+)/)
	{
	    $conRes=$1;
	}
	$i++;
	my @a=@{$clr{$k}};
	my %ID=();
	my %alg=();
	#re-order by seq ID to consensus bsr
	my $n=0;

	my %ligands=();
	my $size = 0;
	`rm -f ligands.pdb`;
	foreach my $line(@a)
	{	    
	    if($line =~ /(\S+)\t(\S+)\t(\S+).+:(\S+)/)
	    {
		my $alg1 = $1;
		my $pdb  = $2;
		my $lig  = $3;
		my $res  = $4;

		$size++;
		
		if($alg1 ne "CON")
		{
		    my $ligname=$lig;
		    if($alg1 ne "FIN")
		    {
			$ligname=substr($lig, 5);
		    }
		    
		    if(exists $ligands{$ligname})
		    {
			$ligands{$ligname}++;
		    }
		    else
		    {
			$ligands{$ligname}=1;
		    }
		}

		my $ligfile="";
		if($alg1 eq "COF")
		{
		    $ligfile="lig_$pdb\_$lig.pdb"
		}
		elsif($alg1 eq "TMS")
		{
		    $ligfile="tms_$pdb\_$lig.pdb"
		}
		elsif($alg1 eq "FIN")
		{
		    $ligfile="lig_$lig.pdb"
		}

		if($ligfile ne "" && -s $ligfile)
		{
		    `cat $ligfile >>ligands.pdb`;
		}
		my $id   = &compare($conRes, $res);
		$ID{$n}  = $id;
		$alg{$n}  = $alg1;
		$n++;
	    }
	}

	if(-s "ligands.pdb")
	{
	    `cat query_tmp.pdb ligands.pdb >CH_complex$i.pdb`;
	}	
	
	

	if($i>1)
	{
	    print OUT1 "\n";
	}
	print OUT1 "SITE $i\t$cscore{$k}\t$size\n";



	my @b=sort {$ID{$b}<=>$ID{$a}} keys %ID;		

	my %reps = ("COF"=>"NA", "TMS"=>"NA", "SST"=>"NA", "FIN"=>"NA", "CON"=>"NA");
	my %repsid = ("COF"=>0, "TMS"=>0, "SST"=>0, "FIN"=>0, "CON"=>0);
	print OUT "cluster $i\n";	
	foreach my $j(@b)
	{	    
	    print OUT "$a[$j]\n";	    
	    my $alg1=$alg{$j};
	    if($reps{$alg1} eq "NA")
	    {
		$reps{$alg1}=$a[$j];
		$repsid{$alg1}=$ID{$j};
	    }
	}

        if($reps{"TMS"} ne "NA")
        {
            print OUT1 $reps{"TMS"} . "\n";

            ##make complex structure
            if($reps{"TMS"} =~ /\S+\t(\S+)\t(\S+)\t.+\t(\S+)\s+:/)
            {
                `cat query_tmp.pdb tms_$1\_$2.pdb >CH_TM_$1\_$2.pdb`;
            }
            else
            {
                print "wrong pattern in TM-SITE results\n";
            }
        }
	elsif($reps{"COF"} ne "NA")
	{
	    print OUT1 $reps{"COF"} . "\n";
	    ##make complex structure
	    if($reps{"COF"} =~ /\S+\t(\S+)\t(\S+)\t.+\t(\S+)\s+:/)
	    {
		`cat query_tmp.pdb lig_$1\_$2.pdb >CH_CF_$1\_$2.pdb`;
	    }
	    else
	    {
		print "wrong pattern in COFACTOR results\n";
	    }
	    
	}
        elsif($reps{"SST"} ne "NA")
        {
            print OUT1 $reps{"SST"} . "\n";
        }
	elsif($reps{"FIN"} ne "NA")
	{
	    print OUT1 $reps{"FIN"} . "\n";
	    ##make complex structure
	    if($reps{"FIN"} =~ /\S+\t(\S+)\t(\S+)\t.+\t(\S+)\s+:/)
	    {
		`cat query_tmp.pdb lig_$2.pdb >CH_FN_$2.pdb`;
	    }
	    else
	    {
		print "wrong pattern in FINDSITE results\n";
	    }
	}
	elsif($reps{"CON"} ne "NA")
	{
	    print OUT1 $reps{"CON"} . "\n";
	}


	my @hets=sort { $ligands{$b}<=>$ligands{$a} } keys %ligands;
	for(my $j=0; $j<@hets; $j++)
	{
	    my $h=$hets[$j];
	    
	    if($j==(@hets-1))
	    {
		print OUT1 "$h($ligands{$h})\n";
	    }
	    else
	    {
		print OUT1 "$h($ligands{$h}),";
	    }
	}
	last if($i>=20);
    }
    close(OUT);
    close(OUT1);

    open(OUT, ">Bsites_prob.dat");
    foreach my $k(@keys)
    {
	print OUT "$prob{$k}\n";
    }
    close(OUT); 

    `mv query_tmp.pdb CH_protein.pdb`;
    `cp -f CH_*.pdb $CHoutdir`;

}








sub dist
{
    my ($x1, $y1, $z1, $x2, $y2, $z2)=@_;
    my $dist =sqrt(($x1 - $x2)**2 + ($y1 - $y2)**2 + ($z1 - $z2)**2);
    return($dist);
}


sub dis_matrix
{
    my($center_file) = @_;

    if(!-s "$center_file")
    {
	print "$center_file is missed\n";
	exit;
    }


    
    my $i=0;
    my %center=();
    open(IN, $center_file);
    while(my $line = <IN>)
    {
	if($line =~ /\t/)
	{	    
	    my @a=split(/\t/, $line);
	    my @xyz=split(/\s+/, $a[3]);
	    $center{$i, 1} = $xyz[0];
	    $center{$i, 2} = $xyz[1];
	    $center{$i, 3} = $xyz[2];
	    $i++;
	}
    }    
    close(IN);

    open(DIS, ">dis.dat");       
    my $n=$i;
    for($i=0; $i<$n; $i++)
    {
	my $x=$center{$i, 1};
	my $y=$center{$i, 2};
	my $z=$center{$i, 3};
	for(my $j=0; $j<$n; $j++)
	{
            if($i==$j)
            {
                $dis[$i][$j]=0;
            }
            elsif($j<$i)
            {
                $dis[$i][$j]=$dis[$j][$i];
            }
            else
            {
		my $x1=$center{$j, 1};
		my $y1=$center{$j, 2};
		my $z1=$center{$j, 3};	    
		$dis[$i][$j] = &dist($x, $y, $z, $x1, $y1, $z1);
	    }
	    printf DIS "%.3f ", $dis[$i][$j];
	}
	print DIS "\n";
    }
    close(DIS);

    return $n;
}


sub load_Ca
{    
    my($dir, $pdbfile) = @_;

    if(!-s "$dir/$pdbfile")
    {
	print "PDB file is missed\n";
	exit;
    }

    `cp $dir/$pdbfile .`;
    
    open(QUE, "$pdbfile");
    my %xyz=();
    my $i=0;
    my %resatoms=();
    
    while(my $line = <QUE>)
    {
	if($line =~ /^ATOM/)
	{
	    my $atom_name=substr($line, 12, 4);
	    if($atom_name eq ' CA ' || $atom_name eq '  CA' || $atom_name eq 'CA  ')	    
	    {		    
		my $res_no = substr($line, 22, 5); $res_no =~ s/\s+//g;
		my $check  = $res_no .  $atom_name;
		if(!exists $resatoms{$check}) #ignore alternative location of atoms
		{
		    $resatoms{$check}=1;

		    $xyz{$res_no, 1} = substr($line, 30, 8);
		    $xyz{$res_no, 2} = substr($line, 38, 8);
		    $xyz{$res_no, 3} = substr($line, 46, 8);
		}
	    }
	}
    }
    close(QUE);

    return %xyz;
}

#the predicted BSR centers by COFACTOR
sub cof_centers
{
    my($dir, $inffile, $outfile) = @_;    

    if(!-s "$dir/$inffile")
    {
	print "information file is missed\n";
	return;
    }

    open(OUT, ">$outfile");

    `cp $dir/$inffile inf_tmp.dat`;
    open(INF, "inf_tmp.dat");
    while(my $line=<INF>)
    {
	if($line=~/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\:(\S+)/)
	{
	    my $hits        = $1;
	    my $TMscore     = $2;
	    my $RMSDa       = $3;
	    my $IDEN        = $4;
	    my $Cov         = $5;
	    my $ligand_name = $6;
	    my $ligands     = $7;
	    my $BS_score    = $8;
	    my $LTM         = $9;
	    my $BS_ID       = $10;
	    my $BS_cov      = $11;
	    my $BS_err      = $12;
	    my $BS_ID1      = $13;
	    my $BS_ID2      = $14;
	    my $PBS         = $15;

	    my $Rank_score =  $TMscore + 2.5*$IDEN + 0.25 * $BS_score;
	    my @res=split(/,/, $PBS);
	    my $x=0; my $y=0; my $z=0; 
	    foreach my $r(@res)
	    {
		$x += $queryCA{$r, 1};
		$y += $queryCA{$r, 2};
		$z += $queryCA{$r, 3};
	    }
	    if(@res>0)
	    {
		$x = $x/@res;
		$y = $y/@res;
		$z = $z/@res;
		printf OUT "COF\t$hits\t$ligands\_$ligand_name\t%.3f %.3f %.3f\t%.2f :$PBS\n", $x, $y, $z, $BS_score;
	    }	    
	}
    }
    close(INF);
    close(OUT);



    `rm inf_tmp.dat`;    
}


#the predicted BSR centers by TM-SITE
sub tms_centers
{
    my($dir, $inffile, $outfile) = @_;    

    if(!-s "$dir/$inffile")
    {
	print "information file is missed\n";
	return;
    }

    open(OUT, ">>$outfile");


    `cp $dir/$inffile inf_tmp.dat`;
    open(INF, "inf_tmp.dat");
    while(my $line=<INF>)
    {
	if($line=~/(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s\S+\s+(.+)/)
	{
	    my $hits        = $1;
	    my $ligand_name = $2;
	    my $Rank_score  = $3;
	    my $PBS         = $4; $PBS =~ s/\s+/,/g;	    

	    my $pdb=substr($hits, 0, 5);
	    my $bsn=substr($hits, 6, 4);	    
	    my @res=split(/,/, $PBS);
	    

	    my $x=0; my $y=0; my $z=0; 
	    foreach my $r(@res)
	    {
		$x += $queryCA{$r, 1};
		$y += $queryCA{$r, 2};
		$z += $queryCA{$r, 3};
	    }
	    if(@res>0)
	    {
		$x = $x/@res;
		$y = $y/@res;
		$z = $z/@res;
		printf OUT "TMS\t$pdb\t$bsn\_$ligand_name\t%.3f %.3f %.3f\t%.2f :$PBS\n", $x, $y, $z, $Rank_score;
	    }	    
	}
    }
    close(INF);
    close(OUT);

    `rm inf_tmp.dat`;    
}


#the predicted BSR centers by S-SITE
sub sst_centers
{
    my($dir, $inffile, $outfile) = @_;    

    if(!-s "$dir/$inffile")
    {
	print "information file is missed\n";
	return;
    }

    open(OUT, ">>$outfile");


    `cp $dir/$inffile inf_tmp.dat`;
    open(INF, "inf_tmp.dat");
    while(my $line=<INF>)
    {
	if($line=~/(\S+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s\S+\s+(.+)/)
	{
	    my $hits        = $1;
	    my $Rank_score  = $2;
	    my $PBS         = $3; $PBS =~ s/\s+/,/g;	    

	    my $pdb=substr($hits, 0, 5);
	    my $bsn=substr($hits, 6);	    
	    my @res=split(/,/, $PBS);
	    

	    my $x=0; my $y=0; my $z=0; 
	    foreach my $r(@res)
	    {
		$x += $queryCA{$r, 1};
		$y += $queryCA{$r, 2};
		$z += $queryCA{$r, 3};
	    }
	    if(@res>0)
	    {
		$x = $x/@res;
		$y = $y/@res;
		$z = $z/@res;
		printf OUT "SST\t$pdb\t$bsn\t%.3f %.3f %.3f\t%.2f :$PBS\n", $x, $y, $z, $Rank_score;
	    }	    
	}
    }
    close(INF);
    close(OUT);

    `rm inf_tmp.dat`;    
}

sub fin_centers
{
    my ($dir, $resultfile, $outname)=@_;

    if(!-s "$dir/$resultfile")
    {
	print "FINDSITE result file is missed\n";
	return;
    }

    open(OUT, ">>$outname");

    my @ligall=`cat $dir/$resultfile`;
    foreach my $line(@ligall)
    {
	chomp($line);
	print OUT "$line\n";	    	    	        
    }    
    close(OUT);
}

#the predicted BSR centers by ConCavity
sub con_centers
{
    my($dir, $inffile, $outfile) = @_;

    if(!-s "$dir/$inffile")
    {
	print "information file is missed\n";
	return;
    }

    open(OUT, ">>$outfile");


    `cp $dir/$inffile inf_tmp.dat`;
    open(INF, "inf_tmp.dat");
    while(my $line=<INF>)
    {
	if($line=~/(\S+)\t(.+)/)
	{
	    my $Rank_score  = $1;
	    my $PBS         = $2; $PBS =~ s/\s+/,/g;	    

	    my @res=split(/,/, $PBS);	    

	    my $x=0; my $y=0; my $z=0; 
	    foreach my $r(@res)
	    {
		$x += $queryCA{$r, 1};
		$y += $queryCA{$r, 2};
		$z += $queryCA{$r, 3};
	    }
	    if(@res>0)
	    {
		$x = $x/@res;
		$y = $y/@res;
		$z = $z/@res;
		printf OUT "CON\tNTAPP\tNTAPP\t%.3f %.3f %.3f\t%.2f :$PBS\n", $x, $y, $z, $Rank_score;
	    }	    
	}
    }
    close(INF);
    close(OUT);

    `rm inf_tmp.dat`;    
}


sub compare
{
    my ($a_ref, $b_ref)=@_;
    my @a=split(/,/, $a_ref);
    my @b=split(/,/, $b_ref);
    my %a1=();
    foreach my $r(@a) 
    {
	$a1{$r}=1;
    }
    
    my $n=0;
    foreach my $r(@b)
    {
	if(exists $a1{$r})
	{
	    $n++;
	}
    }    

    my $id=$n/@a;
    my $id1=$n/@b;
    if($id>$id1)
    {
	$id=$id1;
    }
    
    return $id;
}


