#! /usr/bin/perl -w
use strict;


#########################################################################
#Binding site prediction based on global superpose
#Input: a query structure, modle or native is ok
#Output: the possible binding site of the query protein
#
#The library used is the new library I constructed (exclude DIMER lib)
#########################################################################


`hostname`=~/(\S+)/;
my $node=$1;
my $datetime=`date`;
my $path=`pwd`;

print "Hostname: $node\n";
print "Started : $datetime\n";
print "Path    : $path\n";


my $home   = "/nfs/amino-home";


#=pod
my $user        = "zhanglabs";
my $datadir     = "/nfs/amino-home/zhanglabs/IonCom/output";
my $name        = "51";
my $pdbname     = "protein.pdb";
my $tag         = "TM_51_protein";
my $bindir      = "/nfs/amino-home/zhanglabs/IonCom/bin/COACH/bin";
my $run         = "benchmark";
#=cut

=pod
$user        = "zhanglabs";
$datadir     = "/home/zhanglabs/COACH/output";
$name        = "CH000001";
$pdbname     = "protein.pdb";
$tag         = "TM_CH000001_protein";
$bindir      = "/home/zhanglabs/COACH/bin";
$run         = "real";
=cut

#Set binding site library here
my $BSLdir           = "$home/yangji/library1/BPOCKET";
my $dir_pocket_lib   = "$BSLdir/pocket";
my $libbsr           = "BSITES.bsr";
my $libbsrc          = "BSITES.center";
my $recseqfile       = "BSITES.fasta";

my $LIGdir           = "$home/yangji/library1/COFACTOR/ligand";

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Step 0: parameters set ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
#Set program parameters here
my $nTemplates     = 1;           #how many top templates to select
my $nTemplates1    = 10;          #if not enough templates satisfying tmscoreCutoff, how many top templates to use
my $no_res_cut     = 5;           #minimum number of residues in pocket for using TM-align >=5
my $vote_cut       = 0.25;        #vote for inferring final binding (just for Bsites.dat)
my $cluster_cutoff = 4.0;         #cutoff for average linkage clustering of superposed template ligands
my $quality_cut    = 0.57;        #cutoff for overal quality
my $beta           = 0.5;

my $modelname=$pdbname; $modelname=~s/\.pdb//;
my $result_file = "Temlate_$modelname.score"; ##searched TMsocre between the query and pokcet templates
my $inputdir    = "$datadir/$name";
my $outdir      = "$inputdir/$modelname/tmsite";

my $conpred     = "$modelname.bsr";
my $jsd_score   = "JSD_concavity.dat";

my %Qtemplate=();
#####################check exisistence of necessary file########################
if(!-s "$BSLdir/$recseqfile")
{
    die "$BSLdir/$recseqfile is missed!\n";
}

if(!-s "$inputdir/$pdbname")
{
    die "$inputdir/$pdbname is missed!\n";
}

if(!-s "$BSLdir/$libbsr")
{
    die "$libbsr is missed!\n";
}

if(!-s "$BSLdir/$libbsrc")
{
    die "$libbsrc is missed!\n";
}

if(!-s "$BSLdir/pocket.tar.bz2")
{
    die "pocket.tar.bz2 is missed!\n";
}


my %AA=(
        'ALA'=>'A', 'ARG'=>'R', 'ASN'=>'N', 'ASP'=>'D', 'CYS'=>'C', 'GLN'=>'Q', 'GLU'=>'E', 'GLY'=>'G',
        'HIS'=>'H', 'ILE'=>'I', 'LEU'=>'L', 'LYS'=>'K', 'MET'=>'M', 'PHE'=>'F', 'PRO'=>'P', 'SER'=>'S',
        'THR'=>'T', 'TRP'=>'W', 'TYR'=>'Y', 'VAL'=>'V', 'ASX'=>'N', 'GLX'=>'Q', 'UNK'=>'G',
        'A'=>'ALA', 'R'=>'ARG', 'N'=>'ASN', 'D'=>'ASP', 'C'=>'CYS', 'Q'=>'GLN', 'E'=>'GLU', 'G'=>'GLY',
        'H'=>'HIS', 'I'=>'ILE', 'L'=>'LEU', 'K'=>'LYS', 'M'=>'MET', 'F'=>'PHE', 'P'=>'PRO', 'S'=>'SER',
        'T'=>'THR', 'W'=>'TRP', 'Y'=>'TYR', 'V'=>'VAL', 'B'=>'ASX', 'Z'=>'GLX', 'X'=>'UNK');




###############Temporary directory for working ################################
my $workdir  = "/tmp/$user/$tag";
`/bin/mkdir -p $workdir`;
`/bin/rm -rf $workdir/*`;
chdir "$workdir";


#################copy pakage, input data and library to working dir ############
my $query = "query.pdb";  #name used by pakage

`/bin/cp $bindir/tmsite .`; 
`/bin/cp $bindir/clusteringLig .`; 
`/bin/cp $bindir/NWalign_simple .`; 
`cp $inputdir/$pdbname $query`;

`cp $BSLdir/$libbsr .`;
`cp $BSLdir/$libbsrc .`;
`cp $BSLdir/$recseqfile .`;

`cp $query submitted.pdb`;
&make_ss($workdir, $query);

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Step 0-1: load data ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
my %bslib=(); #all binding sites for all templates in the library
my %bslibd=(); #date
my %bslibc=(); #center
my %bsligname=();
open(LIB, "$libbsr");
while(my $line = <LIB>)
{
    chomp($line);
    if(substr($line, 11, 1) eq 'P')
    {
	last;
    }

    if($line=~/(\S+)\s+\:(.+)\:(\S+)\:(\d+)\:(\S+)/)
    {
	my $rec =$1;
	my $ligname=$2;
	my $bsno=$3;
	my $bres=$5;
	my $site=$rec . "_" . $bsno;
	$bslib{$site}=$bres;
	$ligname =~ s/\s+//g;
	$bsligname{$site}=$ligname;
    }
}
close(LIB);




open(LIB, "$libbsrc");
while(my $line = <LIB>)
{
    chomp($line);
    if($line=~/(\S+):(\S+)/)
    {
	$bslibc{$1}=$2;
    }
}
close(LIB);


#load receptor sequence
my %recseq=();
open(REC, "$recseqfile");
while(my $line = <REC>)
{
    chomp($line);
    if($line =~ /^>(\S+)/)
    {
	my $rec=$1;
	$line = <REC>;
	chomp($line);
	$recseq{$rec}=$line;
    }
}
close(REC);

#load query sequence
my $queryseq=&getfasta($query);
open(FSEQ, ">$inputdir/seq.txt");
print FSEQ ">seq\n";
print FSEQ "$queryseq\n";
close(FSEQ);


#load query structure for only once
my %ss=();
&load_ss($query);

my %xyz=&load_Ca($query);

##make output dir if it does not exist
if(!-d $outdir)
{
    `/bin/mkdir -p $outdir`;
}

############### run ConCavity to extract fragments from query structure ###################

if(!-s "$inputdir/$modelname/concavity/$conpred" || !-s "$inputdir/$jsd_score")
{
    die "please run concavity first\n";
    print "running ConCavity...\n";
}
else
{
    print "copy ConCavity directly...\n";
    `/bin/cp $inputdir/$modelname/concavity/$conpred .`;
    `/bin/cp $inputdir/$jsd_score .`;
}


############## extract fragments from query structure based on ConCavity #################
my %profile=(); 
open(PROFILE,"$conpred") || print "ERROR: $conpred is missing\n";
while(my $line=<PROFILE>)
{   
    if($line=~/^(\d+)/)
    {
	if($line=~/(\S+)\s+(\S+)\s+(\S+)/)
	{	
	    $profile{$1}=$3;	                
	}
    }
}
close(PROFILE);

my %zscore=();
my @frag=();
my $profilecheck=0;
&z_scores(\%profile, \%zscore);


my $ccut=0.0;
my @keys0 = keys %profile;
my @keys00= sort {$a <=> $b} @keys0;
foreach my $k00 (@keys00)
{
    #if($zscore{$k00}>$ccut)
    if($profile{$k00}>$ccut)
    {
	$frag[$profilecheck]=$k00;
	$profilecheck++;
    }
}
if($profilecheck<5)
{
    print "$profilecheck is too small\n";
    #use JSD
}
else
{
    #extract fragment now based on z-score
    #find first residue and last residue
    my $n = keys %ss; 
    my $first=$frag[0];
    my $i=$first;
    while($i>1) #extend to N-terminus
    {
	if($ss{$i} eq 'C') #do not extend Coil
	{
	    last;
	}
	
	if($ss{$i} ne $ss{$i-1})
	{
	    last;
	}
	
	if($ss{$i} eq $ss{$i-1})
	{
	    $i--;
	    $first=$i;
	}
    }

    my $last=$frag[$profilecheck-1];
    $i=$last;
    while($i<$n)#extend to C-terminus
    {
	if($ss{$i} eq 'C')
	{
	    last;
	}

	if($ss{$i} ne $ss{$i+1})
	{
	    last;
	}
	
	if($ss{$i} eq $ss{$i+1})
	{
	    $i++;
	    $last=$i;
	}
    }

    &make_frag($query, $first, $last);
}








#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Step 1: TMalign search ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
#=== Step 1.1: TMalign search on template library --->
#check if this has been done
if(-s "$inputdir/$result_file")
{
    printf "$result_file has been done, copy directly!\n";
    `cp $inputdir/$result_file .`;
}
else
{

    `cp $BSLdir/pocket.tar.bz2 .`;
    print "unzipping pocket library..\n";
    `tar -jxvf pocket.tar.bz2`;    
    `rm -fr pocket.tar.bz2`;
    $dir_pocket_lib="pocket";
    `touch $dir_pocket_lib/*`; ##change the timestamp of the files to avoid being cleared

    printf "Running TM-align to search library...\n";
    &template_search($dir_pocket_lib, $query, $run, $result_file);
    `cp $result_file $inputdir/`;
}



#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Step 2: collect templates  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
my $q_cut=$quality_cut;

my %quality=();
my %pbsr=();
my %pcenter=();
my $relieve_flag=0;
 


print "select templates...\n";
&select_templates($result_file);

my $n_selected = keys %quality;
if($n_selected < $nTemplates)
{
    %quality=();
    $relieve_flag=1;
    print "selected number of templates are too small ($n_selected), use top $nTemplates1\n";        	
    &select_templates1($result_file);
}    
    
#sort templates by average score
my @keys = reverse sort{$quality{$a} <=> $quality{$b}} keys %quality;
$n_selected=@keys;
if($n_selected<1)
{
    die "No templates selected!\n";
}


my $final_templates="selected_templates.lst";
my %final_rec=();
open(FOUT, ">$final_templates");

foreach my $k(@keys)
{            	
    #for the selected templates, if there are templates with quality >0.85, then only use good templates
    my @residues=split(/,/, $pbsr{$k});
    print FOUT "$k $pcenter{$k} @residues\n";
    
}
close(FOUT);


#exit;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Step 3: Clustering  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
printf "Clustering...\n";
`./clusteringLig $final_templates $query $vote_cut $cluster_cutoff`;


my %rot=();

&reRank();


#get ligands
&getligands();

`cp $final_templates $outdir/`;
`cp Bsites* $outdir/`;
`cp pocket.dat $outdir/`;









print "Ending...\n";


#~~~~~~~~~~~~~~~~~~~~~~~ End procedure  ~~~~~~~~~~~~~~~~~~~~~~~~~#

my $time=`date`;
printf "ending time: $time";
`sync`;
`sync`;
sleep(1);


`rm -fr $workdir`;
print "Done!\n";

sub getligands
{
    my $n=0;
    my %bsr=();
    my %cs=();
    open(IN, "Bsites.dat");    
    while(my $line=<IN>)
    {
	if($line=~/(\S+)\t\S+\t\S+\t\S+\t(\S+)/)
	{
	    $n++;
	    $cs{$n}=$1;
	    $bsr{$n}=$2;
	}
    }
    close(IN);

    open(SEL, ">templates.inf");


    my %clr=();
    my @tem=();
    my $pseq="";
    my $complex="";
    my $id0=0;
    my $qstr0=0;
    my $t0="";
    my $p0="";
    $n=0;
    open(IN, "Bsites.clr");    
    open(OUT, ">Bsites_lig.clr");
    open(OUT1, ">Bsites_single.dat");
    while(my $line=<IN>)
    {
	chomp($line);
	if($line=~/cluster/)
	{
	    $pseq="";	    
	    $n++;
	    if($n>1)
	    {
		print COM "END\n";
		close(COM);
		`cp $complex $outdir/`;
		$complex =~ s/.pdb//;

		my $tlig=substr($t0, 0, 10);
		`cat submitted.pdb $tlig.pdb >$complex\_$t0.pdb`;
		`cp $complex\_$t0.pdb $outdir/`;

		my $m=$n-1;
		print OUT1 "$m\t$cs{$m}\t$qstr0\t$t0\t$p0\t";


		my @keys = sort {$clr{$b}<=>$clr{$a}} keys %clr;
		print OUT "Ligand counts in this cluster\n";
		foreach my $k(@keys)
		{
		    print OUT "$k $clr{$k}\n";
		    print OUT1 "$k $clr{$k} ";
		}
		print OUT "\n\n";
		print OUT1 "\n";
	    }
	    print OUT "Cluster $n\n";
	    $id0=0;
	    $qstr0=0;
	    %clr=();
	    $t0="";
	    
	    my $pbsr=$bsr{$n};
	    my @a=split(/,/, $pbsr);
	    foreach my $r(@a)
	    {
		$r = $r-1;
		$pseq .= substr($queryseq, $r, 1);
	    }
	    #print "$pbsr\t$pseq\n";

	    $line=<IN>;
	
	    
	    if($line =~ /(\S+)\s+\S+\s+\S+\s+\S+\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(.+)/)
	    {
		my $t = $1;
		my $q = $2;
		my $p = $3;
		my $lig=$bsligname{$t};
		if(exists $clr{$lig})
		{
		    $clr{$lig}++;		    
		}
		else
		{
		    $clr{$lig}=1;
		}
		my $temp=substr($line, 11);
		if($temp =~ /\S+\s+\S+\s+\S+\s+(.+)/)
		{
		    print OUT "$t $lig " . $1 . "\n";
		    print SEL "$t $lig " . $1 . "\n";
		}

		my @matrix=split(/,/, $rot{$t});
		my @rot=();
		my @tra=();
		$tra[0]    = $matrix[0];
		$rot[0][0] = $matrix[1];
		$rot[0][1] = $matrix[2];
		$rot[0][2] = $matrix[3];

		$tra[1]    = $matrix[4];
		$rot[1][0] = $matrix[5];
		$rot[1][1] = $matrix[6];
		$rot[1][2] = $matrix[7];
		
		$tra[2]    = $matrix[8];
		$rot[2][0] = $matrix[9];
		$rot[2][1] = $matrix[10];
		$rot[2][2] = $matrix[11];
		
		
		
		my @a=split(/\s+/, $p);
		my $seq="";
		foreach my $r(@a)
		{
		    $r = $r-1;
		    $seq .= substr($queryseq, $r, 1);
		}
		#print "$t\t$lig\t$q\t$p\t$seq\n";

		if(!-s "$LIGdir/$t.pdb")
		{
		    die "$LIGdir/$t.pdb is missed\n";
		}
		
		my @lig=`cat $LIGdir/$t.pdb`;
		open(TMP, ">$t.pdb");
		open(LIGS, ">tms_$t\_$lig.pdb");
		print TMP "TER\n";
		$complex="complex$n.pdb";
		`cp submitted.pdb $complex`;
		open(COM, ">>$complex");
		print COM "TER consensus binding residues: $bsr{$n}\n";		
		foreach my $r(@lig)
		{	
		    chomp($r);
		    if($r !~ /^HETATM/)
		    {
			next;
		    }

		    my $x = substr($r, 30, 8);
		    my $y = substr($r, 38, 8);
		    my $z = substr($r, 46, 8);
		    
		    my $rx = $rot[0][0]*$x + $rot[0][1]*$y + $rot[0][2]*$z + $tra[0];
		    my $ry = $rot[1][0]*$x + $rot[1][1]*$y + $rot[1][2]*$z + $tra[1];
		    my $rz = $rot[2][0]*$x + $rot[2][1]*$y + $rot[2][2]*$z + $tra[2];

		    substr($r, 30, 8)=sprintf("%8.3f", $rx);
		    substr($r, 38, 8)=sprintf("%8.3f", $ry);
		    substr($r, 46, 8)=sprintf("%8.3f", $rz);
		    print COM "$r\n";
		    print LIGS "$r\n";
		    print TMP "$r\n";
		}
		close(TMP);
		

		my $id=&compare($bsr{$n}, $p);
		printf COM "TER $t $lig q_str=$q, Seq id to consensus=%.2f\n", $id;
		$p =~ s/\s+/,/g;
		print COM "TER binding residues: $p\n";
		printf LIGS "TER TM-SITE\t$t $lig:$p\n";
		close(LIGS);
		if($id>$id0)
		{
		    $id0=$id;
		    $qstr0=$q;
		    $t0="$t\_$lig";
		    $p0=$p;
		}
		elsif($id==$id0)
		{
		    if($qstr0>$q)
		    {
			$qstr0=$q;
			$t0="$t\_$lig";
			$p0=$p;
		    }
		}
	    }
	}
	else
	{
	    if($line =~ /(\S+)\s+\S+\s+\S+\s+\S+\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(.+)/)
	    {
		my $t = $1;
		my $q = $2;
		my $p = $3;
		my $lig=$bsligname{$t};

		my $temp=substr($line, 11);
		if($temp =~ /\S+\s+\S+\s+\S+\s+(.+)/)
		{
		    print OUT "$t $lig " . $1 . "\n";
		    print SEL "$t $lig " . $1 . "\n";
		}

		if(exists $clr{$lig})
		{
		    $clr{$lig}++;		    
		}
		else
		{
		    $clr{$lig}=1;		    
		}

		my @matrix=split(/,/, $rot{$t});
		my @rot=();
		my @tra=();
		$tra[0]    = $matrix[0];
		$rot[0][0] = $matrix[1];
		$rot[0][1] = $matrix[2];
		$rot[0][2] = $matrix[3];
		
		$tra[1]    = $matrix[4];
		$rot[1][0] = $matrix[5];
		$rot[1][1] = $matrix[6];
		$rot[1][2] = $matrix[7];
		
		$tra[2]    = $matrix[8];
		$rot[2][0] = $matrix[9];
		$rot[2][1] = $matrix[10];
		$rot[2][2] = $matrix[11];		    
		
		my @lig=`cat $LIGdir/$t.pdb`;
		open(TMP, ">$t.pdb");
		open(LIGS, ">tms_$t\_$lig.pdb");
		print TMP "TER\n";
		foreach my $r(@lig)
		{	
		    chomp($r);
		    if($r !~ /^HETATM/)
		    {
			next;
		    }
		    
		    my $x = substr($r, 30, 8);
		    my $y = substr($r, 38, 8);
		    my $z = substr($r, 46, 8);
		    
		    my $rx = $rot[0][0]*$x + $rot[0][1]*$y + $rot[0][2]*$z + $tra[0];
		    my $ry = $rot[1][0]*$x + $rot[1][1]*$y + $rot[1][2]*$z + $tra[1];
		    my $rz = $rot[2][0]*$x + $rot[2][1]*$y + $rot[2][2]*$z + $tra[2];
		    
		    substr($r, 30, 8)=sprintf("%8.3f", $rx);
		    substr($r, 38, 8)=sprintf("%8.3f", $ry);
		    substr($r, 46, 8)=sprintf("%8.3f", $rz);
		    print COM "$r\n";
		    print LIGS "$r\n";
		    print TMP "$r\n";
		}
		close(TMP);
		my $id=&compare($bsr{$n}, $p);
		printf COM "TER $t $lig q_str=$q, Seq id to consensus=%.2f\n", $id;
		$p =~ s/\s+/,/g;
		print COM "TER binding residues: $p\n";
		printf LIGS "TER TM-SITE\t$t $lig:$p\n";
		close(LIGS);
		
		if($id>$id0)
		{
		    $id0=$id;
		    $qstr0=$q;
		    $t0="$t\_$lig";
		    $p0=$p;
		}
		elsif($id==$id0)
		{
		    if($qstr0>$q)
		    {
			$qstr0=$q;
			$t0="$t\_$lig";
			$p0=$p;
		    }
		}		
		#print "$lig $line\n";
	    }
	}	
    }
    close(IN);
    

    if($n>0)
    {
	print COM "END\n";
	close(COM);
	`cp $complex $outdir/`;
	$complex =~ s/.pdb//;
	
	my $tlig=substr($t0, 0, 10);
	`cat submitted.pdb $tlig.pdb >$complex\_$t0.pdb`;
	`cp $complex\_$t0.pdb $outdir/`;
	my $m=$n;
	print OUT1 "$m\t$cs{$m}\t$qstr0\t$t0\t$p0\t";


	my @keys = sort {$clr{$b}<=>$clr{$a}} keys %clr;
	print OUT "ligand count\n";
	foreach my $k(@keys)
	{
	    print OUT "$k $clr{$k}\n";
	    print OUT1 "$k $clr{$k} ";
	}
	print OUT "\n\n";
	print OUT1 "\n";	
    }

    close(OUT);
    close(OUT1);
    close(SEL);

    `cp tms_*.pdb $outdir/`;
    `cp templates.inf $outdir/`;
}

sub compare
{
    my ($a_ref, $b_ref)=@_;
    my @a=split(/,/, $a_ref);
    my @b=split(/\s+/, $b_ref);
    my %a1=();
    foreach my $r(@a) 
    {
	$a1{$r}=1;
    }
    
    my $n=0;
    foreach my $r(@b)
    {
	if(exists $a1{$r})
	{
	    $n++;
	}
    }    

    my $id=$n/@a;
    my $id1=$n/@b;
    if($id>$id1)
    {
	$id=$id1;
    }
    
    return $id;
}

sub reRank
{
    #re-rank Bsites.dat. Bsites,clr, Bsites_prob.dat based on c-score

    my %quality=();
    my %pred=();
    my $len=0;
    my $i=0;

    my @prob=();
   
    open(FIN, "$jsd_score");
    my %JSD=();
    while(my $line = <FIN>)
    {
        if($line =~ /(\d+)\s+\S+\s+(\S+)/)
        {
            my $res=$1+1;
            my $score=$2;
            $JSD{$res}=$score;
        }
    }
    close(FIN);

    

    open(IN, "Bsites_prob.dat");
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /\d+/)
	{
	    push(@prob, $line);
	    my @l=split(/\s+/, $line);
	    $len=@l;
	}
    }
    close(IN);

    
    my @conser=();
    $i=0;
    open(IN, "Bsites.dat");
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /(\S+):(\S+)/)
	{
	    my $s=$1;
	    $quality{$i}= $s;
	    $pred{$i}=$2;
	    my @consensus0=split(/,/, $pred{$i});
	    
	    $conser[$i]=0;
	    foreach my $c(@consensus0)
	    {
		$conser[$i] += $JSD{$c};
	    }	    
	    $conser[$i] = $conser[$i]/@consensus0;
	    $i++;
	}
    }
    close(IN);


    my %clr=();
    my @ntemp=();
    my @highestq=();
    my $T=0;
    my %consensus=();
    
    $i=-1;
    my $sum=0;
    my $count=0;
    my $h=0;
    my %selected=();
    open(IN, "Bsites.clr");    
    while(my $line=<IN>)
    {
	chomp($line);
	if($line =~ /cluster/)
	{
	    if($i>=0)
	    {
		$T+=$ntemp[$i];
		$highestq[$i]=$h;
	    }
	    $i++;
	    $ntemp[$i]=0;
	    %consensus=();
	    $sum=0;
	    $h=0;
	    $count=0;
	    my @consensus0=split(/,/, $pred{$i});
	    foreach my $c(@consensus0)
	    {
		$consensus{$c}=1;
	    }
	}
	else
	{	 
	    push(@{$clr{$i}}, $line);
	    $ntemp[$i]++;


	    if($line=~/(\S+)\s+\S+\s+\S+\s+\S+\s+(\S+)/)
	    {
		my $site=$1;
		my $q=$2;
		$selected{$site}=1;
		if($q>$h)
		{
		    $h=$q;
		}
	    }
	}		
    }
    close(IN);
    if($i>=0)
    {
	$T+=$ntemp[$i];
	$highestq[$i]=$h;
    }

   
    my @keys=  keys %quality;
    my %cscore=();
    foreach my $k(@keys)
    {
	my $size  = $ntemp[$k]/$T;
	my $size1 = log(1+$ntemp[$k] ** (1.0/3));
	$cscore{$k} = 2/(1+exp( -$size*$highestq[$k] - 0.2*$size1 - 0.2*$conser[$k] )  ) -1;
    }


    @keys= sort {$cscore{$b} <=> $cscore{$a}} keys %cscore;
    

    open(OUT, ">Bsites.dat");
    foreach my $k(@keys)
    {
	printf OUT "%.2f\t%.2f\t%.2f\t%.2f\t$pred{$k}\n", $cscore{$k}, $highestq[$k], $conser[$k], $ntemp[$k]/$T;
    }
    close(OUT); 

    
    open(TM, $result_file) or die "$result_file is missed\n";
    while(my $line=<TM>)
    {
	chomp($line);
	if($line =~ /:/)
	{
	    my @a=split(/:/, $line);
	    if(exists $selected{$a[0]})
	    {
		$rot{$a[0]}=$a[3];		
	    }
	}
    }
    close(TM);
	

    open(OUT, ">Bsites.clr");
    foreach my $k(@keys)
    {
	my @a=@{$clr{$k}};
	my %quality=();
	
	$i=0;
	foreach my $p(@a)
	{
	    if($p=~/\S+\s+\S+\s+\S+\s+\S+\s+(\S+)/)
	    {
		my $q=$1;
		$quality{$i}=$q;
		$i++;
	    }
	}

	my @c = sort {$quality{$b} <=> $quality{$a}} keys %quality;
	print OUT "cluster $k\n";
	foreach my $j(@c)
	{
	    print OUT "$a[$j]\n";
	}
    }
    close(OUT);



    open(OUT, ">Bsites_prob.dat");
    foreach my $k(@keys)
    {
	print OUT "$prob[$k]\n";
    }
    close(OUT); 

}


sub load_ss
{
    my ($pdbfile)=@_;
	
    open(QUE, "<$pdbfile") or die "$pdbfile file is missed!\n";
    my @res_no=();
    my $i=0;
    while(my $line = <QUE>)
    {
	if($line =~ /^ATOM/)
	{
	    my $alt=substr($line, 16, 1); #alterate for atom
	    if($alt eq " " || $alt eq "A"|| $alt eq "1")
	    {
		my $str=substr($line, 12, 2); 		
		if($str =~ /H/g)
		{
		    next;
		}
		
		my $n = substr($line, 22, 4);  $n =~ s/\s+//;
		$res_no[$i]=$n;
		$i++;

		if(!exists $ss{$n})
		{
		    $ss{$n}=substr($line, 79, 1);
		}
	    }
	}
    }
    close(QUE);
}

sub load_Ca
{
    my ($pdbfile)=@_;
	
    open(QUE, "<$pdbfile") or die "$pdbfile file is missed!\n";
    my %xyz=();
    my $i=0;
    my %resatoms=();
    
    while(my $line = <QUE>)
    {
	if($line =~ /^ATOM/)
	{
	    my $atom_name=substr($line, 12, 4);
	    if($atom_name eq ' CA ' || $atom_name eq '  CA' || $atom_name eq 'CA  ')
	    {		    
		my $res_no = substr($line, 22, 5); $res_no =~ s/\s+//g;
		my $check  = $res_no .  $atom_name;
		if(!exists $resatoms{$check}) #ignore alternative location of atoms
		{
		    $resatoms{$check}=1;

		    $xyz{$res_no, 1} = substr($line, 30, 8);
		    $xyz{$res_no, 2} = substr($line, 38, 8);
		    $xyz{$res_no, 3} = substr($line, 46, 8);
		}
	    }
	}
    }
    close(QUE);

    return %xyz;
}





sub dist
{
    my $dist=0;
    my @xyz=();
    $xyz[0]=$_[0];
    $xyz[1]=$_[1];
    $xyz[2]=$_[2];

    my @xyz1=();    
    $xyz1[0]=$_[3];
    $xyz1[1]=$_[4];
    $xyz1[2]=$_[5];
    
    $dist=sqrt(($xyz[0]-$xyz1[0])**2 + ($xyz[1]-$xyz1[1])**2 + ($xyz[2]-$xyz1[2])**2);
    
    return $dist;
}

sub max
{
    my ($a, $b)=@_;
    if($a>$b)
    {
	return $a;
    }
    else
    {
	return $b;
    }
}


sub min
{
    my ($a, $b)=@_;
    if($a<$b)
    {
	return $a;
    }
    else
    {
	return $b;
    }
}



sub template_search
{
    my ($dir_pocket_lib, $query, $run, $result_file) = @_;
    my $seqid="";
    my %chains=();
    my @keys= keys %bslib;
    open(FRES, ">$result_file");

    foreach my $rec(@keys)
    {
	my $chain=substr($rec, 0, 5);	
	if(!exists $chains{$chain})
	{	    
	    if($run eq "benchmark")
	    {
		my $seq1=$recseq{$chain};   
		$seqid=`./NWalign_simple $seq1 $queryseq 3`;
		$chains{$chain}=$seqid;
		if($seqid>0.3)
		{
		    next;
		}
	    }
	}
	else
	{
	    if($chains{$chain}>0.3)
	    {
		next;
	    }
	}
	
	my $line=$bslib{$rec};		
	my @pocres=split(/,/, $line);
	my $nres=$#pocres;
	my $firstRes=sprintf "%d", $pocres[0];
	my $lastRes=sprintf "%d", $pocres[$nres];

	if(!-s "$dir_pocket_lib/$rec.pdb") {die "$dir_pocket_lib/$rec.pdb is missed\n";}
	if(($lastRes-$firstRes)<=$no_res_cut) 
	{
	    #check the pocket file to see if it was extended
	    my $tseq=&getfasta("$dir_pocket_lib/$rec.pdb");
	    if(length $tseq<=$no_res_cut)
	    {
		next;
	    }
	}	    
	print FRES "$rec";
	    
	my $cmd="./tmsite -B $dir_pocket_lib/$rec.pdb -A $query -rB $line";
	#print "$cmd\n";
	my @rst=`$cmd`;
	#print "@rst";
	my $i=0;
	my $flag=0;
	my $predBSR="";
	my $Lg=0;
	my $Ls=0;
	my $Lcov=0;
	for($i=0; $i<=$#rst; $i++)
	{
	    my $r=$rst[$i];
            if($r =~ /^predicted BSR:(\S+)/)
            {
                $predBSR=$1;
            }
            if($r =~ /^Lg=(\S+)/)
            {
                $Lg=$1;
            }
            if($r =~ /^Ls=(\S+)/)
            {
                $Ls=$1;
            }
           if($r =~ /^Lcov=(\S+)/)
            {
                $Lcov=$1;
            }

	    if($r =~ /^TM-score=(\S+)\s+/)
	    {
		printf FRES ":%6.5f", $1;
	    }
	    if($rst[$i]=~/\s+1\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/)
	    {	
		printf FRES ":%.5f,%.5f,%.5f,%.5f", $1,$2,$3,$4;	
	    }
	    if($rst[$i]=~/\s+2\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/)
	    {
		printf FRES ",%.5f,%.5f,%.5f,%.5f", $1,$2,$3,$4;
	    }
	    if($rst[$i]=~/\s+3\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/)
	    {
		printf FRES ",%.5f,%.5f,%.5f,%.5f", $1,$2,$3,$4;
		$flag=1;
		last;
	    }
	}
	print FRES ":$Lg:$Ls:$Lcov:$predBSR";
        if($flag==0)
        {
            print "wrong running of TM-align for $rec\n";
            print FRES "ERROR";
        }

	print FRES "\n";
    }
    close(FRES);
}



sub getfasta
{
    my $seq="";
    my ($pdb)=@_;
    open(FPDB, $pdb) or die "$pdb is missed!\n";
    my %resatoms=();
    while(my $line = <FPDB>)
    {
	if($line =~ /^ATOM/i)
	{
            my $res_name  = substr($line, 17, 3);  
            my $res_no    = substr($line, 22, 5);
            my $check     = $res_no .  substr($line, 12, 4); #atom

	    if(!exists $resatoms{$check}) ##check for alternative position atoms
	    {
		$resatoms{$check}=1;

		my $atom_name=substr($line, 12, 4);
		if($atom_name eq ' CA ' || $atom_name eq '  CA' || $atom_name eq 'CA  ')
		{
		    $seq = $seq. $AA{$res_name};
		}		
	    }
	}
    }
    close(FPDB);

    return $seq;
}




sub select_templates
{
    my ($filename)=@_;   


    open(FIN, "$jsd_score") or die "No jsdscore\n";
    my %JSD=();
    while(my $line = <FIN>)
    {
        if($line =~ /(\d+)+\s+\S+\s+(\S+)/)
        {
            my $res=$1+1;
            my $score=$2;
            $JSD{$res}=$score;
        }
    }
    close(FIN);


    my @tx=();
    my @ty=();
    my @tz=();
    my @tatom=();
    my %t_res_no=();
    my @bs_template=();
    my $n=0;


    open(FIN, $filename) or die "$filename is missed!\n";    
    while(my $line1 = <FIN>)
    {
	chomp($line1);
        if($line1 =~ /ERROR/)
        {
            next;
        }

	if($line1 =~ /(\S+)/)
	{	   	    
	    my @inf=split(/:/, $line1, -1);

	    my $site=$inf[0];
	    #if($bslibd{$site} == 1)
	    #{
		#next;
	    #}
	    my $pdb=substr($site, 0, 5);
	    #if($pdb ne "3luoA") {next;}
	    my $bres=$bslib{$site};    	   

	    my $score1=$inf[1]; #tmscore normlized by query
	    my $score2=$inf[2]; #tmscore normlized by template
	    my @matrix=split(/,/, $inf[3]);
	    my @rot=();
	    my @tra=();
	    $tra[0]    = $matrix[0];
	    $rot[0][0] = $matrix[1];
	    $rot[0][1] = $matrix[2];
	    $rot[0][2] = $matrix[3];

	    $tra[1]    = $matrix[4];
	    $rot[1][0] = $matrix[5];
	    $rot[1][1] = $matrix[6];
	    $rot[1][2] = $matrix[7];

	    $tra[2]    = $matrix[8];
	    $rot[2][0] = $matrix[9];
	    $rot[2][1] = $matrix[10];
	    $rot[2][2] = $matrix[11];

	    my $Lg      = $inf[4];
	    my $Ls      = $inf[5];
	    my $Lcov    = $inf[6];
	    my $predBSR = $inf[7];

	    if($predBSR !~ /\d+/) 
	    {
		#print "ignore $site $predBSR\n";
		next;
	    }
	    my @pred=split(/,/, $predBSR);
	    my $cons=0;
	    foreach my $p(@pred)
	    {
		if(!exists $JSD{$p})
		{
		    print "wrong res $p\n";
		    next;
		}
		$cons += $JSD{$p};
	    }
	    if(@pred>0)
	    {
		$cons=$cons/@pred;
	    }

	    my $tm=2.0*($score1*$score2)/($score1+$score2);

	    my $q = 2.0/(1.0 + exp( -( $Lcov*(0.4*$Lg+0.3*$Ls+0.2*$cons) +  $tm)**2) )-1.0;
	    $Qtemplate{$site}=$q;
    
	    if($q>$q_cut)
	    {   
		my @pBSR=split(/,/, $predBSR);

		if(!exists $bslibc{$site})
		{
		    die "$site\n";
		}
		#my @tc=split(/,/, $bslibc{$site});

		my $x=0; my $y=0; my $z=0; my $wrong=0;
		foreach my $b(@pBSR)
		{
		    if(!exists $xyz{$b, 1})
		    {
			print "wrong residue $b\n";
			$wrong=1;
			last;
		    }

		    $x += $xyz{$b, 1};
		    $y += $xyz{$b, 2};
		    $z += $xyz{$b, 3};
		}
		$x = $x/@pBSR;
		$y = $y/@pBSR;
		$z = $z/@pBSR;
		
		if($wrong==1)
		{
		    next;
		}

		#my $rx=$rot[0][0]*$tc[0]+$rot[0][1]*$tc[1]+$rot[0][2]*$tc[2]+$tra[0];
		#my $ry=$rot[1][0]*$tc[0]+$rot[1][1]*$tc[1]+$rot[1][2]*$tc[2]+$tra[1];
		#my $rz=$rot[2][0]*$tc[0]+$rot[2][1]*$tc[1]+$rot[2][2]*$tc[2]+$tra[2];
	    
		$quality{$site} = $q;
		$pbsr{$site}    = $predBSR;
		#$pcenter{$site} = sprintf "%.5f %.5f %.5f %.2f %.2f %.2f %.2f %.2f %.2f", $rx, $ry, $rz, $q, $score1, $score2, $Lg, $Ls, $Lcov;
		$pcenter{$site} = sprintf "%.5f %.5f %.5f %.2f %.2f %.2f %.2f %.2f %.2f", $x, $y, $z, $q, $score1, $score2, $Lg, $Ls, $Lcov;
	    }
	}
    }
    close(FIN);
}


sub select_templates1
{
    my ($filename)=@_;

    my @tx=();
    my @ty=();
    my @tz=();
    my @tatom=();
    my %t_res_no=();
    my @bs_template=();
    my $n=0;

    my @allt = sort {$Qtemplate{$b} <=> $Qtemplate{$a}} keys %Qtemplate;


    my %topT=();
    for(my $i=0; $i<$nTemplates1; $i++)
    {
	$topT{$allt[$i]}=1;
    }

    open(FIN, $filename) or die "$filename is missed!\n";    
    while(my $line1 = <FIN>)
    {
	chomp($line1);
        if($line1 =~ /ERROR/)
        {
            next;
        }

	if($line1 =~ /(\S+)/)
	{	    
	    my @inf=split(/:/, $line1);

	    my $site=$inf[0];
	    #if($bslibd{$site} == 1)
	    #{
		#next;
	    #}

	    if(!exists $topT{$site})
	    {
		next;
	    }
	    my $pdb=substr($site, 0, 5);
	    #if($pdb ne "3luoA") {next;}
	    my $bres=$bslib{$site};    	   

	    my $score1=$inf[1]; #tmscore normlized by query
	    my $score2=$inf[2]; #tmscore normlized by template
	    my @matrix=split(/,/, $inf[3]);
	    my @rot=();
	    my @tra=();
	    $tra[0]    = $matrix[0];
	    $rot[0][0] = $matrix[1];
	    $rot[0][1] = $matrix[2];
	    $rot[0][2] = $matrix[3];

	    $tra[1]    = $matrix[4];
	    $rot[1][0] = $matrix[5];
	    $rot[1][1] = $matrix[6];
	    $rot[1][2] = $matrix[7];

	    $tra[2]    = $matrix[8];
	    $rot[2][0] = $matrix[9];
	    $rot[2][1] = $matrix[10];
	    $rot[2][2] = $matrix[11];

	    my $Lg      = $inf[4];
	    my $Ls      = $inf[5];
	    my $Lcov    = $inf[6];
	    my $predBSR = $inf[7];

	    my $q=$Qtemplate{$site};   


	    my @pBSR=split(/,/, $predBSR);
	    
	    if(!exists $bslibc{$site})
	    {
		die "$site\n";
	    }
	    #my @tc=split(/,/, $bslibc{$site});
	    #my $rx=$rot[0][0]*$tc[0]+$rot[0][1]*$tc[1]+$rot[0][2]*$tc[2]+$tra[0];
	    #my $ry=$rot[1][0]*$tc[0]+$rot[1][1]*$tc[1]+$rot[1][2]*$tc[2]+$tra[1];
	    #my $rz=$rot[2][0]*$tc[0]+$rot[2][1]*$tc[1]+$rot[2][2]*$tc[2]+$tra[2];


	    my $x=0; my $y=0; my $z=0; my $wrong=0;
	    foreach my $b(@pBSR)
	    {
		if(!exists $xyz{$b, 1})
		{
		    print "wrong residue $b\n";
		    $wrong=1;
		    last;
		}
		
		$x += $xyz{$b, 1};
		$y += $xyz{$b, 2};
		$z += $xyz{$b, 3};
	    }
	    $x = $x/@pBSR;
	    $y = $y/@pBSR;
	    $z = $z/@pBSR;
	    
	    if($wrong==1)
	    {
		next;
	    }

	    
	    $quality{$site} = $q;
	    $pbsr{$site}    = $predBSR;
	    #$pcenter{$site} = sprintf "%.5f %.5f %.5f %.2f %.2f %.2f %.2f %.2f %.2f", $rx, $ry, $rz, $q, $score1, $score2, $Lg, $Ls, $Lcov;	    
	    $pcenter{$site} = sprintf "%.5f %.5f %.5f %.2f %.2f %.2f %.2f %.2f %.2f", $x, $y, $z, $q, $score1, $score2, $Lg, $Ls, $Lcov;
	}
    }
    close(FIN);
}




sub z_scores
{
     my($score_ref, $z_score_ref)=@_;
     my %r_score  =%$score_ref;

     my @keys = keys %r_score;     
     my $Lch = @keys;
     my $z1_a  = 0; my $z1_a2=0; my %zscore=();
     foreach my $i(@keys)
     {
         $z1_a  += $r_score{$i};
         $z1_a2 += $r_score{$i}**2;
     }
     $z1_a/=$Lch;
     $z1_a2/=$Lch;
     my $z1_sd = sqrt($z1_a2 - $z1_a**2);

     foreach my $i(@keys)
     {
         $$z_score_ref{$i}=($r_score{$i}-$z1_a)/$z1_sd;
     }
}


sub make_frag
{
    my ($pdbfile, $first, $last)=@_;
	
    open(QUE, "<$pdbfile") or die "$pdbfile file is missed!\n";
    open(FQUE, ">qfrag.pdb");
    while(my $line = <QUE>)
    {
	if($line =~ /^ATOM/)
	{
	    my $alt=substr($line, 16, 1); #alterate for atom
	    if($alt eq " " || $alt eq "A"|| $alt eq "1")
	    {
		my $resno=sprintf "%d", substr($line, 22, 4);
		
		if($resno>=$first && $resno<=$last)
		{
		    print FQUE "$line";
		}		
	    }
	}
    }
    close(QUE);
    close(FQUE);

    `cp qfrag.pdb $outdir/`;
    `cp $pdbfile query0.pdb`;
    `mv qfrag.pdb $pdbfile`; #change original model to fragments
}

sub PMCC
{
    #Pseduo MCC
    my %native = %{$_[0]};
    my @pred = @{$_[1]};
    my $n_all_res=$_[2];

    my @nat=keys %native;


    my $TP=0; my $FP=0; my $TN=0; my $FN=0;
    foreach my $p(@pred)
    {
	$p =~ s/\s+//g;
	if(exists $native{$p})
	{
	    $TP++;
	}
	else
	{
	    $FP++;
	}
    }
    $FN=@nat-$TP;
    $TN=$n_all_res-($FN+$TP+$FP);
    my $demoinator=sqrt(($TP+$FP)*($TP+$FN)*($TN+$FP)*($TN+$FN));
    #printf "$TP, $TN, $FP, $FN\n";
    
    my $mcc = 0;
    my $acc = 0;
    my $cov = 0;

    if($demoinator==0)
    {
	$mcc = 0;
    }
    else
    {
	$mcc = ($TP*$TN - $FP*$FN)/$demoinator;
    }
   
    return $mcc;	
}



sub make_ss
{
    my ($dir, $model)=@_;
    
    my %HEC=('H'=>'H', 'G'=>'H', 'I'=>'H', 'B'=>'E', 'b'=>'E', 'E'=>'E', 'T'=>'C', 'C'=>'C');
    my @rec=();
    
    #generate SS with stride if the model is accurate
    #otherwise, use PSIPRED to predict SS ?? currently, this is not done
    open(FREC, "<$dir/$model");
    while(my $line1 = <FREC>)
    {
        chomp($line1);
        push(@rec, $line1);
    }
    close(FREC);

    my %ss=();
    my @rst=`$bindir/stride $dir/$model`;
    foreach my $r(@rst)
    {
        if($r =~ /^ASG/)
        {
            my $resno=sprintf "%d", substr($r, 11, 4); #original residue number
            my $s=$HEC{substr($r, 24, 1)};
            $ss{$resno}=$s;
        }
    }

    open(FREC, ">$dir/$model"); #overwrite the protein
    foreach my $r(@rec)
    {
        if($r =~ /^ATOM/)
        {
            my $resno=sprintf "%d", substr($r, 22, 4);
            my $s="C";
            if(exists $ss{$resno})
            {
                $s=$ss{$resno};
            }
            my $len=length $r;
            my $r1=substr($r, 0, $len);
            for(my $i=$len; $i<80; $i++)
            {
                $r1 .= " ";
            }

            substr($r1, 79, 1)= sprintf "%1s", $s;
            substr($r1, 55, 1) =sprintf "%1s", $s;

            my $atom=substr($r1, 12, 2);
            if($atom =~ /H/)
            {
                next;
            }
            print FREC "$r1\n";
        }
        else
        {
            print FREC "$r\n";
        }
    }
    close(FREC);
}


